#ifndef KITEVM_H
#define KITEVM_H

#ifndef NO_INCLUDE_IOSTREAM
#include <iostream>
#endif
#include <string>
#include <stack>
#include "KiteReference.h"
#include "SymbolTable.h"

using namespace std;

class KiteObject;

extern string processEscapes(string str);

namespace KiteVM {
	class Instruction;
    
    class FuncCall {
    public:
        FuncCall() { }
        
        FuncCall(string name, string file, int line)
        : name(name), file(file), line(line) { }
        
        FuncCall(const FuncCall &obj)
        : name(obj.name), file(obj.file), line(obj.line) { }
    
        string name;
        string file;
        int line;
    };
    
	class VM {
	public:
		VM();
		virtual ~VM();

		KiteReference<KiteObject> run(Instruction *instr);
	//private:
		stack<KiteReference<KiteObject> > stack; // This _shouldn't_ be public.
        vector<FuncCall> funcCalls;
		SymbolTable syms;
	};

	class Instruction {
	public:
		Instruction(std::string file, int line);
		virtual ~Instruction();

		void addToEnd(Instruction *p);
		virtual KiteReference<KiteObject> eval(VM &vm) = 0;
		virtual std::string toString() const = 0;

		void deallocate();
		void outTree();

		friend class VM;
	protected:
		Instruction *next;
        std::string file;
        int line;
	};

	class SingleArgumentInstruction : public Instruction {
	public:
		SingleArgumentInstruction(std::string file, int line, Instruction *arg1) : Instruction(file, line) {
			arg1->addToEnd(this);
		}

		virtual KiteReference<KiteObject> eval(VM &vm) = 0;
	};

	class LoadInstruction : public SingleArgumentInstruction {
	public:
		LoadInstruction(std::string file, int line, Instruction *arg1) : SingleArgumentInstruction(file, line, arg1) { }
		virtual std::string toString() const { return "LOAD"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class DefinitionInstruction : public Instruction {
	public:
		DefinitionInstruction(std::string file, int line, Instruction *lhs, Instruction *rhs) : Instruction(file, line) {
			rhs->addToEnd(lhs);
			lhs->addToEnd(this);
		}

		virtual KiteReference<KiteObject> eval(VM &vm) = 0;
	};

	class VarInstruction : public DefinitionInstruction {
	public:
		VarInstruction(std::string file, int line, Instruction *lhs, Instruction *rhs) : DefinitionInstruction(file, line, lhs, rhs) { }
		virtual std::string toString() const { return "DEFINEVAR"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class ClassInstruction : public DefinitionInstruction {
	public:
		ClassInstruction(std::string file, int line, Instruction *lhs, Instruction *rhs) : DefinitionInstruction(file, line, lhs, rhs) { }
		virtual std::string toString() const { return "DEFINECLASS"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class FunctionCallInstruction : public Instruction {
	public:
		FunctionCallInstruction(std::string file, int line, Instruction *object, Instruction *name, Instruction *arguments) : Instruction(file, line) {
			if (arguments) arguments->addToEnd(name);
			name->addToEnd(object);
			object->addToEnd(this);
		}

		virtual std::string toString() const { return "CALL"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	template<class T>
	class ConstantInstruction : public Instruction {
	public:
		ConstantInstruction(std::string file, int line, T val)
			: Instruction(file, line), constant(val) { }

		virtual KiteReference<KiteObject> eval(VM &vm) = 0;
	protected:
		T constant;
	};

	class StringInstruction : public ConstantInstruction<std::string> {
	public:
		StringInstruction(std::string file, int line, const char *string) : ConstantInstruction<std::string>(file, line, processEscapes(std::string(string))) { }
		virtual std::string toString() const { return "STRING " + constant; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class IdentifierInstruction : public ConstantInstruction<std::string> {
	public:
		IdentifierInstruction(std::string file, int line, const char *string) : ConstantInstruction<std::string>(file, line, std::string(string)) { }
		virtual std::string toString() const { return "IDENTIFIER " + constant; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class NumberInstruction : public ConstantInstruction<std::string> {
	public:
		NumberInstruction(std::string file, int line, const char *string) : ConstantInstruction<std::string>(file, line, std::string(string)) { }
		virtual std::string toString() const { return "NUMBER " + constant; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class BooleanInstruction : public ConstantInstruction<std::string> {
	public:
		BooleanInstruction(std::string file, int line, bool val) : ConstantInstruction<std::string>(file, line, val ? std::string("true") : std::string("false")) { }
		virtual std::string toString() const { return "BOOLEAN " + constant; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class CodeInstruction : public ConstantInstruction<Instruction*> {
	public:
		CodeInstruction(std::string file, int line, Instruction *code) : ConstantInstruction<Instruction*>(file, line, code) { }
		virtual ~CodeInstruction() { constant->deallocate(); }
		virtual std::string toString() const { 
			std::string ret = "CODE {\n";
			constant->outTree();
			return ret + "\n}";
		}

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class AssociateArrayInstruction : public DefinitionInstruction {
	public:
		AssociateArrayInstruction(std::string file, int line, Instruction *lhs, Instruction *rhs) : DefinitionInstruction(file, line, lhs, rhs) { }
		virtual std::string toString() const { return "ASSOCIATE"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class AssignInstruction : public DefinitionInstruction {
	public:
		AssignInstruction(std::string file, int line, Instruction *lhs, Instruction *rhs) : DefinitionInstruction(file, line, lhs, rhs) { }
		virtual std::string toString() const { return "ASSIGN"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class ArrayDereferenceInstruction : public DefinitionInstruction {
	public:
		ArrayDereferenceInstruction(std::string file, int line, Instruction *lhs, Instruction *rhs) : DefinitionInstruction(file, line, lhs, rhs) { }
		virtual std::string toString() const { return "ARRAYELEMENT"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class EmptyArrayInstruction : public Instruction {
	public:
        EmptyArrayInstruction(std::string file, int line) : Instruction(file, line) { }
		virtual std::string toString() const { return "EMPTYARRAY"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class MergeArrayInstruction : public DefinitionInstruction {
	public:
		MergeArrayInstruction(std::string file, int line, Instruction *lhs, Instruction *rhs) : DefinitionInstruction(file, line, lhs, rhs) { }
		virtual std::string toString() const { return "MERGEARRAY"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class DereferenceInstruction : public DefinitionInstruction {
	public:
		DereferenceInstruction(std::string file, int line, Instruction *lhs, Instruction *rhs) : DefinitionInstruction(file, line, lhs, rhs) { }
		virtual std::string toString() const { return "DEREFERENCE"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};
};

#define THIS_OBJECT() (vm.syms.get("this"))

#include "ParserState.h"

KiteReference<KiteObject> coerceNumber(KiteVM::VM &vm, KiteReference<KiteObject> obj);
KiteReference<KiteObject> coerceString(KiteVM::VM &vm, KiteReference<KiteObject> obj);
bool valuesEqual(KiteVM::VM &vm, KiteReference<KiteObject> x, KiteReference<KiteObject> j);

extern void *ParseAlloc(void *(*mallocProc)(size_t));
extern void Parse(void *yyp, int yymajor, KiteVM::Instruction *yyminor, ParserState *s);
extern void ParseFree(void *p, void (*freeProc)(void*));

#ifdef ALTERNATE_ALLOCATOR
void *operator new(size_t size);
void operator delete(void *ptr, size_t size);
void operator delete(void *ptr);
#endif

#endif // KITEVM_H
